@@ -1,12 +1,10 @@
 import os
 
-from cryptography.hazmat.backends import default_backend
 from cryptography.hazmat.primitives import hashes
 from cryptography.hazmat.primitives.asymmetric import ec
 from cryptography.hazmat.primitives.asymmetric.utils import (
     encode_dss_signature,
 )
-
 from tests.utils import load_fips_ecdsa_signing_vectors, load_vectors_from_file
 
 CRYPTOGRAPHY_HASH_TYPES = {
@@ -27,7 +25,7 @@ def verify_one_vector(vector):
 
     numbers = ec.EllipticCurvePublicNumbers(x, y, ec.SECP256K1())
 
-    key = numbers.public_key(default_backend())
+    key = numbers.public_key()
 
     verifier = key.verifier(
         signature, ec.ECDSA(CRYPTOGRAPHY_HASH_TYPES[digest_algorithm]())
