@@ -2,6 +2,7 @@
 # 2.0, and the BSD License. See the LICENSE file in the root of this repository
 # for complete details.
 
+from __future__ import annotations
 
 import binascii
 import enum
@@ -10,6 +11,7 @@ import re
 import typing
 import warnings
 from base64 import encodebytes as _base64_encode
+from dataclasses import dataclass
 
 from cryptography import utils
 from cryptography.exceptions import UnsupportedAlgorithm
@@ -22,7 +24,12 @@ from cryptography.hazmat.primitives.asymmetric import (
     rsa,
 )
 from cryptography.hazmat.primitives.asymmetric import utils as asym_utils
-from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
+from cryptography.hazmat.primitives.ciphers import (
+    AEADDecryptionContext,
+    Cipher,
+    algorithms,
+    modes,
+)
 from cryptography.hazmat.primitives.serialization import (
     Encoding,
     KeySerializationEncryption,
@@ -57,6 +64,10 @@ _ECDSA_NISTP384 = b"ecdsa-sha2-nistp384"
 _ECDSA_NISTP521 = b"ecdsa-sha2-nistp521"
 _CERT_SUFFIX = b"-cert-v01@openssh.com"
 
+# U2F application string suffixed pubkey
+_SK_SSH_ED25519 = b"sk-ssh-ed25519@openssh.com"
+_SK_SSH_ECDSA_NISTP256 = b"sk-ecdsa-sha2-nistp256@openssh.com"
+
 # These are not key types, only algorithms, so they cannot appear
 # as a public key type
 _SSH_RSA_SHA256 = b"rsa-sha2-256"
@@ -77,18 +88,47 @@ _PEM_RC = re.compile(_SK_START + b"(.*?)" + _SK_END, re.DOTALL)
 # padding for max blocksize
 _PADDING = memoryview(bytearray(range(1, 1 + 16)))
 
+
+@dataclass
+class _SSHCipher:
+    alg: type[algorithms.AES]
+    key_len: int
+    mode: type[modes.CTR] | type[modes.CBC] | type[modes.GCM]
+    block_len: int
+    iv_len: int
+    tag_len: int | None
+    is_aead: bool
+
+
 # ciphers that are actually used in key wrapping
-_SSH_CIPHERS: typing.Dict[
-    bytes,
-    typing.Tuple[
-        typing.Type[algorithms.AES],
-        int,
-        typing.Union[typing.Type[modes.CTR], typing.Type[modes.CBC]],
-        int,
-    ],
-] = {
-    b"aes256-ctr": (algorithms.AES, 32, modes.CTR, 16),
-    b"aes256-cbc": (algorithms.AES, 32, modes.CBC, 16),
+_SSH_CIPHERS: dict[bytes, _SSHCipher] = {
+    b"aes256-ctr": _SSHCipher(
+        alg=algorithms.AES,
+        key_len=32,
+        mode=modes.CTR,
+        block_len=16,
+        iv_len=16,
+        tag_len=None,
+        is_aead=False,
+    ),
+    b"aes256-cbc": _SSHCipher(
+        alg=algorithms.AES,
+        key_len=32,
+        mode=modes.CBC,
+        block_len=16,
+        iv_len=16,
+        tag_len=None,
+        is_aead=False,
+    ),
+    b"aes256-gcm@openssh.com": _SSHCipher(
+        alg=algorithms.AES,
+        key_len=32,
+        mode=modes.GCM,
+        block_len=16,
+        iv_len=12,
+        tag_len=16,
+        is_aead=True,
+    ),
 }
 
 # map local curve name to key type
@@ -99,9 +139,7 @@ _ECDSA_KEY_TYPE = {
 }
 
 
-def _get_ssh_key_type(
-    key: typing.Union["SSHPrivateKeyTypes", "SSHPublicKeyTypes"]
-) -> bytes:
+def _get_ssh_key_type(key: SSHPrivateKeyTypes | SSHPublicKeyTypes) -> bytes:
     if isinstance(key, ec.EllipticCurvePrivateKey):
         key_type = _ecdsa_key_type(key.public_key())
     elif isinstance(key, ec.EllipticCurvePublicKey):
@@ -131,20 +169,20 @@ def _ecdsa_key_type(public_key: ec.EllipticCurvePublicKey) -> bytes:
 
 
 def _ssh_pem_encode(
-    data: bytes,
+    data: utils.Buffer,
     prefix: bytes = _SK_START + b"\n",
     suffix: bytes = _SK_END + b"\n",
 ) -> bytes:
     return b"".join([prefix, _base64_encode(data), suffix])
 
 
-def _check_block_size(data: bytes, block_len: int) -> None:
+def _check_block_size(data: utils.Buffer, block_len: int) -> None:
     """Require data to be full blocks"""
     if not data or len(data) % block_len != 0:
         raise ValueError("Corrupt data: missing padding")
 
 
-def _check_empty(data: bytes) -> None:
+def _check_empty(data: utils.Buffer) -> None:
     """All data should have been parsed."""
     if data:
         raise ValueError("Corrupt data: unparsed data")
@@ -152,34 +190,41 @@ def _check_empty(data: bytes) -> None:
 
 def _init_cipher(
     ciphername: bytes,
-    password: typing.Optional[bytes],
+    password: bytes | None,
     salt: bytes,
     rounds: int,
-) -> Cipher[typing.Union[modes.CBC, modes.CTR]]:
+) -> Cipher[modes.CBC | modes.CTR | modes.GCM]:
     """Generate key + iv and return cipher."""
     if not password:
-        raise ValueError("Key is password-protected.")
+        raise TypeError(
+            "Key is password-protected, but password was not provided."
+        )
 
-    algo, key_len, mode, iv_len = _SSH_CIPHERS[ciphername]
-    seed = _bcrypt_kdf(password, salt, key_len + iv_len, rounds, True)
-    return Cipher(algo(seed[:key_len]), mode(seed[key_len:]))
+    ciph = _SSH_CIPHERS[ciphername]
+    seed = _bcrypt_kdf(
+        password, salt, ciph.key_len + ciph.iv_len, rounds, True
+    )
+    return Cipher(
+        ciph.alg(seed[: ciph.key_len]),
+        ciph.mode(seed[ciph.key_len :]),
+    )
 
 
-def _get_u32(data: memoryview) -> typing.Tuple[int, memoryview]:
+def _get_u32(data: memoryview) -> tuple[int, memoryview]:
     """Uint32"""
     if len(data) < 4:
         raise ValueError("Invalid data")
     return int.from_bytes(data[:4], byteorder="big"), data[4:]
 
 
-def _get_u64(data: memoryview) -> typing.Tuple[int, memoryview]:
+def _get_u64(data: memoryview) -> tuple[int, memoryview]:
     """Uint64"""
     if len(data) < 8:
         raise ValueError("Invalid data")
     return int.from_bytes(data[:8], byteorder="big"), data[8:]
 
 
-def _get_sshstr(data: memoryview) -> typing.Tuple[memoryview, memoryview]:
+def _get_sshstr(data: memoryview) -> tuple[memoryview, memoryview]:
     """Bytes with u32 length prefix"""
     n, data = _get_u32(data)
     if n > len(data):
@@ -187,7 +232,7 @@ def _get_sshstr(data: memoryview) -> typing.Tuple[memoryview, memoryview]:
     return data[:n], data[n:]
 
 
-def _get_mpint(data: memoryview) -> typing.Tuple[int, memoryview]:
+def _get_mpint(data: memoryview) -> tuple[int, memoryview]:
     """Big integer."""
     val, data = _get_sshstr(data)
     if val and val[0] > 0x7F:
@@ -208,16 +253,14 @@ def _to_mpint(val: int) -> bytes:
 class _FragList:
     """Build recursive structure without data copy."""
 
-    flist: typing.List[bytes]
+    flist: list[utils.Buffer]
 
-    def __init__(
-        self, init: typing.Optional[typing.List[bytes]] = None
-    ) -> None:
+    def __init__(self, init: list[utils.Buffer] | None = None) -> None:
         self.flist = []
         if init:
             self.flist.extend(init)
 
-    def put_raw(self, val: bytes) -> None:
+    def put_raw(self, val: utils.Buffer) -> None:
         """Add plain bytes"""
         self.flist.append(val)
 
@@ -229,7 +272,7 @@ class _FragList:
         """Big-endian uint64"""
         self.flist.append(val.to_bytes(length=8, byteorder="big"))
 
-    def put_sshstr(self, val: typing.Union[bytes, "_FragList"]) -> None:
+    def put_sshstr(self, val: bytes | _FragList) -> None:
         """Bytes prefixed with u32 length"""
         if isinstance(val, (bytes, memoryview, bytearray)):
             self.put_u32(len(val))
@@ -270,7 +313,9 @@ class _SSHFormatRSA:
         mpint n, e, d, iqmp, p, q
     """
 
-    def get_public(self, data: memoryview):
+    def get_public(
+        self, data: memoryview
+    ) -> tuple[tuple[int, int], memoryview]:
         """RSA public fields"""
         e, data = _get_mpint(data)
         n, data = _get_mpint(data)
@@ -278,7 +323,7 @@ class _SSHFormatRSA:
 
     def load_public(
         self, data: memoryview
-    ) -> typing.Tuple[rsa.RSAPublicKey, memoryview]:
+    ) -> tuple[rsa.RSAPublicKey, memoryview]:
         """Make RSA public key from data."""
         (e, n), data = self.get_public(data)
         public_numbers = rsa.RSAPublicNumbers(e, n)
@@ -286,8 +331,8 @@ class _SSHFormatRSA:
         return public_key, data
 
     def load_private(
-        self, data: memoryview, pubfields
-    ) -> typing.Tuple[rsa.RSAPrivateKey, memoryview]:
+        self, data: memoryview, pubfields, unsafe_skip_rsa_key_validation: bool
+    ) -> tuple[rsa.RSAPrivateKey, memoryview]:
         """Make RSA private key from data."""
         n, data = _get_mpint(data)
         e, data = _get_mpint(data)
@@ -304,7 +349,9 @@ class _SSHFormatRSA:
         private_numbers = rsa.RSAPrivateNumbers(
             p, q, d, dmp1, dmq1, iqmp, public_numbers
         )
-        private_key = private_numbers.private_key()
+        private_key = private_numbers.private_key(
+            unsafe_skip_rsa_key_validation=unsafe_skip_rsa_key_validation
+        )
         return private_key, data
 
     def encode_public(
@@ -340,9 +387,7 @@ class _SSHFormatDSA:
         mpint p, q, g, y, x
     """
 
-    def get_public(
-        self, data: memoryview
-    ) -> typing.Tuple[typing.Tuple, memoryview]:
+    def get_public(self, data: memoryview) -> tuple[tuple, memoryview]:
         """DSA public fields"""
         p, data = _get_mpint(data)
         q, data = _get_mpint(data)
@@ -352,7 +397,7 @@ class _SSHFormatDSA:
 
     def load_public(
         self, data: memoryview
-    ) -> typing.Tuple[dsa.DSAPublicKey, memoryview]:
+    ) -> tuple[dsa.DSAPublicKey, memoryview]:
         """Make DSA public key from data."""
         (p, q, g, y), data = self.get_public(data)
         parameter_numbers = dsa.DSAParameterNumbers(p, q, g)
@@ -362,8 +407,8 @@ class _SSHFormatDSA:
         return public_key, data
 
     def load_private(
-        self, data: memoryview, pubfields
-    ) -> typing.Tuple[dsa.DSAPrivateKey, memoryview]:
+        self, data: memoryview, pubfields, unsafe_skip_rsa_key_validation: bool
+    ) -> tuple[dsa.DSAPrivateKey, memoryview]:
         """Make DSA private key from data."""
         (p, q, g, y), data = self.get_public(data)
         x, data = _get_mpint(data)
@@ -421,7 +466,7 @@ class _SSHFormatECDSA:
 
     def get_public(
         self, data: memoryview
-    ) -> typing.Tuple[typing.Tuple, memoryview]:
+    ) -> tuple[tuple[memoryview, memoryview], memoryview]:
         """ECDSA public fields"""
         curve, data = _get_sshstr(data)
         point, data = _get_sshstr(data)
@@ -433,17 +478,17 @@ class _SSHFormatECDSA:
 
     def load_public(
         self, data: memoryview
-    ) -> typing.Tuple[ec.EllipticCurvePublicKey, memoryview]:
+    ) -> tuple[ec.EllipticCurvePublicKey, memoryview]:
         """Make ECDSA public key from data."""
-        (curve_name, point), data = self.get_public(data)
+        (_, point), data = self.get_public(data)
         public_key = ec.EllipticCurvePublicKey.from_encoded_point(
             self.curve, point.tobytes()
         )
         return public_key, data
 
     def load_private(
-        self, data: memoryview, pubfields
-    ) -> typing.Tuple[ec.EllipticCurvePrivateKey, memoryview]:
+        self, data: memoryview, pubfields, unsafe_skip_rsa_key_validation: bool
+    ) -> tuple[ec.EllipticCurvePrivateKey, memoryview]:
         """Make ECDSA private key from data."""
         (curve_name, point), data = self.get_public(data)
         secret, data = _get_mpint(data)
@@ -486,14 +531,14 @@ class _SSHFormatEd25519:
 
     def get_public(
         self, data: memoryview
-    ) -> typing.Tuple[typing.Tuple, memoryview]:
+    ) -> tuple[tuple[memoryview], memoryview]:
         """Ed25519 public fields"""
         point, data = _get_sshstr(data)
         return (point,), data
 
     def load_public(
         self, data: memoryview
-    ) -> typing.Tuple[ed25519.Ed25519PublicKey, memoryview]:
+    ) -> tuple[ed25519.Ed25519PublicKey, memoryview]:
         """Make Ed25519 public key from data."""
         (point,), data = self.get_public(data)
         public_key = ed25519.Ed25519PublicKey.from_public_bytes(
@@ -502,8 +547,8 @@ class _SSHFormatEd25519:
         return public_key, data
 
     def load_private(
-        self, data: memoryview, pubfields
-    ) -> typing.Tuple[ed25519.Ed25519PrivateKey, memoryview]:
+        self, data: memoryview, pubfields, unsafe_skip_rsa_key_validation: bool
+    ) -> tuple[ed25519.Ed25519PrivateKey, memoryview]:
         """Make Ed25519 private key from data."""
         (point,), data = self.get_public(data)
         keypair, data = _get_sshstr(data)
@@ -541,6 +586,70 @@ class _SSHFormatEd25519:
         f_priv.put_sshstr(f_keypair)
 
 
+def load_application(data) -> tuple[memoryview, memoryview]:
+    """
+    U2F application strings
+    """
+    application, data = _get_sshstr(data)
+    if not application.tobytes().startswith(b"ssh:"):
+        raise ValueError(
+            "U2F application string does not start with b'ssh:' "
+            f"({application})"
+        )
+    return application, data
+
+
+class _SSHFormatSKEd25519:
+    """
+    The format of a sk-ssh-ed25519@openssh.com public key is:
+
+        string		"sk-ssh-ed25519@openssh.com"
+        string		public key
+        string		application (user-specified, but typically "ssh:")
+    """
+
+    def load_public(
+        self, data: memoryview
+    ) -> tuple[ed25519.Ed25519PublicKey, memoryview]:
+        """Make Ed25519 public key from data."""
+        public_key, data = _lookup_kformat(_SSH_ED25519).load_public(data)
+        _, data = load_application(data)
+        return public_key, data
+
+    def get_public(self, data: memoryview) -> typing.NoReturn:
+        # Confusingly `get_public` is an entry point used by private key
+        # loading.
+        raise UnsupportedAlgorithm(
+            "sk-ssh-ed25519 private keys cannot be loaded"
+        )
+
+
+class _SSHFormatSKECDSA:
+    """
+    The format of a sk-ecdsa-sha2-nistp256@openssh.com public key is:
+
+        string		"sk-ecdsa-sha2-nistp256@openssh.com"
+        string		curve name
+        ec_point	Q
+        string		application (user-specified, but typically "ssh:")
+    """
+
+    def load_public(
+        self, data: memoryview
+    ) -> tuple[ec.EllipticCurvePublicKey, memoryview]:
+        """Make ECDSA public key from data."""
+        public_key, data = _lookup_kformat(_ECDSA_NISTP256).load_public(data)
+        _, data = load_application(data)
+        return public_key, data
+
+    def get_public(self, data: memoryview) -> typing.NoReturn:
+        # Confusingly `get_public` is an entry point used by private key
+        # loading.
+        raise UnsupportedAlgorithm(
+            "sk-ecdsa-sha2-nistp256 private keys cannot be loaded"
+        )
+
+
 _KEY_FORMATS = {
     _SSH_RSA: _SSHFormatRSA(),
     _SSH_DSA: _SSHFormatDSA(),
@@ -548,10 +657,12 @@ _KEY_FORMATS = {
     _ECDSA_NISTP256: _SSHFormatECDSA(b"nistp256", ec.SECP256R1()),
     _ECDSA_NISTP384: _SSHFormatECDSA(b"nistp384", ec.SECP384R1()),
     _ECDSA_NISTP521: _SSHFormatECDSA(b"nistp521", ec.SECP521R1()),
+    _SK_SSH_ED25519: _SSHFormatSKEd25519(),
+    _SK_SSH_ECDSA_NISTP256: _SSHFormatSKECDSA(),
 }
 
 
-def _lookup_kformat(key_type: bytes):
+def _lookup_kformat(key_type: utils.Buffer):
     """Return valid format or throw error"""
     if not isinstance(key_type, bytes):
         key_type = memoryview(key_type).tobytes()
@@ -569,9 +680,11 @@ SSHPrivateKeyTypes = typing.Union[
 
 
 def load_ssh_private_key(
-    data: bytes,
-    password: typing.Optional[bytes],
+    data: utils.Buffer,
+    password: bytes | None,
     backend: typing.Any = None,
+    *,
+    unsafe_skip_rsa_key_validation: bool = False,
 ) -> SSHPrivateKeyTypes:
     """Load private key from OpenSSH custom encoding."""
     utils._check_byteslike("data", data)
@@ -603,11 +716,7 @@ def load_ssh_private_key(
     pubfields, pubdata = kformat.get_public(pubdata)
     _check_empty(pubdata)
 
-    # load secret data
-    edata, data = _get_sshstr(data)
-    _check_empty(data)
-
-    if (ciphername, kdfname) != (_NONE, _NONE):
+    if ciphername != _NONE or kdfname != _NONE:
         ciphername_bytes = ciphername.tobytes()
         if ciphername_bytes not in _SSH_CIPHERS:
             raise UnsupportedAlgorithm(
@@ -615,14 +724,40 @@ def load_ssh_private_key(
             )
         if kdfname != _BCRYPT:
             raise UnsupportedAlgorithm(f"Unsupported KDF: {kdfname!r}")
-        blklen = _SSH_CIPHERS[ciphername_bytes][3]
+        blklen = _SSH_CIPHERS[ciphername_bytes].block_len
+        tag_len = _SSH_CIPHERS[ciphername_bytes].tag_len
+        # load secret data
+        edata, data = _get_sshstr(data)
+        # see https://bugzilla.mindrot.org/show_bug.cgi?id=3553 for
+        # information about how OpenSSH handles AEAD tags
+        if _SSH_CIPHERS[ciphername_bytes].is_aead:
+            tag = bytes(data)
+            if len(tag) != tag_len:
+                raise ValueError("Corrupt data: invalid tag length for cipher")
+        else:
+            _check_empty(data)
         _check_block_size(edata, blklen)
         salt, kbuf = _get_sshstr(kdfoptions)
         rounds, kbuf = _get_u32(kbuf)
         _check_empty(kbuf)
         ciph = _init_cipher(ciphername_bytes, password, salt.tobytes(), rounds)
-        edata = memoryview(ciph.decryptor().update(edata))
+        dec = ciph.decryptor()
+        edata = memoryview(dec.update(edata))
+        if _SSH_CIPHERS[ciphername_bytes].is_aead:
+            assert isinstance(dec, AEADDecryptionContext)
+            _check_empty(dec.finalize_with_tag(tag))
+        else:
+            # _check_block_size requires data to be a full block so there
+            # should be no output from finalize
+            _check_empty(dec.finalize())
     else:
+        if password:
+            raise TypeError(
+                "Password was given but private key is not encrypted."
+            )
+        # load secret data
+        edata, data = _get_sshstr(data)
+        _check_empty(data)
         blklen = 8
         _check_block_size(edata, blklen)
     ck1, edata = _get_u32(edata)
@@ -634,8 +769,13 @@ def load_ssh_private_key(
     key_type, edata = _get_sshstr(edata)
     if key_type != pub_key_type:
         raise ValueError("Corrupt data: key type mismatch")
-    private_key, edata = kformat.load_private(edata, pubfields)
-    comment, edata = _get_sshstr(edata)
+    private_key, edata = kformat.load_private(
+        edata,
+        pubfields,
+        unsafe_skip_rsa_key_validation=unsafe_skip_rsa_key_validation,
+    )
+    # We don't use the comment
+    _, edata = _get_sshstr(edata)
 
     # yes, SSH does padding check *after* all other parsing is done.
     # need to follow as it writes zero-byte padding too.
@@ -675,7 +815,7 @@ def _serialize_ssh_private_key(
     f_kdfoptions = _FragList()
     if password:
         ciphername = _DEFAULT_CIPHER
-        blklen = _SSH_CIPHERS[ciphername][3]
+        blklen = _SSH_CIPHERS[ciphername].block_len
         kdfname = _BCRYPT
         rounds = _DEFAULT_ROUNDS
         if (
@@ -757,11 +897,11 @@ class SSHCertificate:
         _serial: int,
         _cctype: int,
         _key_id: memoryview,
-        _valid_principals: typing.List[bytes],
+        _valid_principals: list[bytes],
         _valid_after: int,
         _valid_before: int,
-        _critical_options: typing.Dict[bytes, bytes],
-        _extensions: typing.Dict[bytes, bytes],
+        _critical_options: dict[bytes, bytes],
+        _extensions: dict[bytes, bytes],
         _sig_type: memoryview,
         _sig_key: memoryview,
         _inner_sig_type: memoryview,
@@ -813,7 +953,7 @@ class SSHCertificate:
         return bytes(self._key_id)
 
     @property
-    def valid_principals(self) -> typing.List[bytes]:
+    def valid_principals(self) -> list[bytes]:
         return self._valid_principals
 
     @property
@@ -825,11 +965,11 @@ class SSHCertificate:
         return self._valid_after
 
     @property
-    def critical_options(self) -> typing.Dict[bytes, bytes]:
+    def critical_options(self) -> dict[bytes, bytes]:
         return self._critical_options
 
     @property
-    def extensions(self) -> typing.Dict[bytes, bytes]:
+    def extensions(self) -> dict[bytes, bytes]:
         return self._extensions
 
     def signature_key(self) -> SSHCertPublicKeyTypes:
@@ -889,9 +1029,9 @@ def _get_ec_hash_alg(curve: ec.EllipticCurve) -> hashes.HashAlgorithm:
 
 
 def _load_ssh_public_identity(
-    data: bytes,
+    data: utils.Buffer,
     _legacy_dsa_allowed=False,
-) -> typing.Union[SSHCertificate, SSHPublicKeyTypes]:
+) -> SSHCertificate | SSHPublicKeyTypes:
     utils._check_byteslike("data", data)
 
     m = _SSH_PUBKEY_RC.match(data)
@@ -984,13 +1124,13 @@ def _load_ssh_public_identity(
 
 
 def load_ssh_public_identity(
-    data: bytes,
-) -> typing.Union[SSHCertificate, SSHPublicKeyTypes]:
+    data: utils.Buffer,
+) -> SSHCertificate | SSHPublicKeyTypes:
     return _load_ssh_public_identity(data)
 
 
-def _parse_exts_opts(exts_opts: memoryview) -> typing.Dict[bytes, bytes]:
-    result: typing.Dict[bytes, bytes] = {}
+def _parse_exts_opts(exts_opts: memoryview) -> dict[bytes, bytes]:
+    result: dict[bytes, bytes] = {}
     last_name = None
     while exts_opts:
         name, exts_opts = _get_sshstr(exts_opts)
@@ -1000,13 +1140,39 @@ def _parse_exts_opts(exts_opts: memoryview) -> typing.Dict[bytes, bytes]:
         if last_name is not None and bname < last_name:
             raise ValueError("Fields not lexically sorted")
         value, exts_opts = _get_sshstr(exts_opts)
+        if len(value) > 0:
+            value, extra = _get_sshstr(value)
+            if len(extra) > 0:
+                raise ValueError("Unexpected extra data after value")
         result[bname] = bytes(value)
         last_name = bname
     return result
 
 
+def ssh_key_fingerprint(
+    key: SSHPublicKeyTypes,
+    hash_algorithm: hashes.MD5 | hashes.SHA256,
+) -> bytes:
+    if not isinstance(hash_algorithm, (hashes.MD5, hashes.SHA256)):
+        raise TypeError("hash_algorithm must be either MD5 or SHA256")
+
+    key_type = _get_ssh_key_type(key)
+    kformat = _lookup_kformat(key_type)
+
+    f_pub = _FragList()
+    f_pub.put_sshstr(key_type)
+    kformat.encode_public(key, f_pub)
+
+    ssh_binary_data = f_pub.tobytes()
+
+    # Hash the binary data
+    hash_obj = hashes.Hash(hash_algorithm)
+    hash_obj.update(ssh_binary_data)
+    return hash_obj.finalize()
+
+
 def load_ssh_public_key(
-    data: bytes, backend: typing.Any = None
+    data: utils.Buffer, backend: typing.Any = None
 ) -> SSHPublicKeyTypes:
     cert_or_key = _load_ssh_public_identity(data, _legacy_dsa_allowed=True)
     public_key: SSHPublicKeyTypes
@@ -1060,16 +1226,16 @@ _SSHKEY_CERT_MAX_PRINCIPALS = 256
 class SSHCertificateBuilder:
     def __init__(
         self,
-        _public_key: typing.Optional[SSHCertPublicKeyTypes] = None,
-        _serial: typing.Optional[int] = None,
-        _type: typing.Optional[SSHCertificateType] = None,
-        _key_id: typing.Optional[bytes] = None,
-        _valid_principals: typing.List[bytes] = [],
+        _public_key: SSHCertPublicKeyTypes | None = None,
+        _serial: int | None = None,
+        _type: SSHCertificateType | None = None,
+        _key_id: bytes | None = None,
+        _valid_principals: list[bytes] = [],
         _valid_for_all_principals: bool = False,
-        _valid_before: typing.Optional[int] = None,
-        _valid_after: typing.Optional[int] = None,
-        _critical_options: typing.List[typing.Tuple[bytes, bytes]] = [],
-        _extensions: typing.List[typing.Tuple[bytes, bytes]] = [],
+        _valid_before: int | None = None,
+        _valid_after: int | None = None,
+        _critical_options: list[tuple[bytes, bytes]] = [],
+        _extensions: list[tuple[bytes, bytes]] = [],
     ):
         self._public_key = _public_key
         self._serial = _serial
@@ -1084,7 +1250,7 @@ class SSHCertificateBuilder:
 
     def public_key(
         self, public_key: SSHCertPublicKeyTypes
-    ) -> "SSHCertificateBuilder":
+    ) -> SSHCertificateBuilder:
         if not isinstance(
             public_key,
             (
@@ -1110,7 +1276,7 @@ class SSHCertificateBuilder:
             _extensions=self._extensions,
         )
 
-    def serial(self, serial: int) -> "SSHCertificateBuilder":
+    def serial(self, serial: int) -> SSHCertificateBuilder:
         if not isinstance(serial, int):
             raise TypeError("serial must be an integer")
         if not 0 <= serial < 2**64:
@@ -1131,7 +1297,7 @@ class SSHCertificateBuilder:
             _extensions=self._extensions,
         )
 
-    def type(self, type: SSHCertificateType) -> "SSHCertificateBuilder":
+    def type(self, type: SSHCertificateType) -> SSHCertificateBuilder:
         if not isinstance(type, SSHCertificateType):
             raise TypeError("type must be an SSHCertificateType")
         if self._type is not None:
@@ -1150,7 +1316,7 @@ class SSHCertificateBuilder:
             _extensions=self._extensions,
         )
 
-    def key_id(self, key_id: bytes) -> "SSHCertificateBuilder":
+    def key_id(self, key_id: bytes) -> SSHCertificateBuilder:
         if not isinstance(key_id, bytes):
             raise TypeError("key_id must be bytes")
         if self._key_id is not None:
@@ -1170,8 +1336,8 @@ class SSHCertificateBuilder:
         )
 
     def valid_principals(
-        self, valid_principals: typing.List[bytes]
-    ) -> "SSHCertificateBuilder":
+        self, valid_principals: list[bytes]
+    ) -> SSHCertificateBuilder:
         if self._valid_for_all_principals:
             raise ValueError(
                 "Principals can't be set because the cert is valid "
@@ -1227,9 +1393,7 @@ class SSHCertificateBuilder:
             _extensions=self._extensions,
         )
 
-    def valid_before(
-        self, valid_before: typing.Union[int, float]
-    ) -> "SSHCertificateBuilder":
+    def valid_before(self, valid_before: int | float) -> SSHCertificateBuilder:
         if not isinstance(valid_before, (int, float)):
             raise TypeError("valid_before must be an int or float")
         valid_before = int(valid_before)
@@ -1251,9 +1415,7 @@ class SSHCertificateBuilder:
             _extensions=self._extensions,
         )
 
-    def valid_after(
-        self, valid_after: typing.Union[int, float]
-    ) -> "SSHCertificateBuilder":
+    def valid_after(self, valid_after: int | float) -> SSHCertificateBuilder:
         if not isinstance(valid_after, (int, float)):
             raise TypeError("valid_after must be an int or float")
         valid_after = int(valid_after)
@@ -1277,7 +1439,7 @@ class SSHCertificateBuilder:
 
     def add_critical_option(
         self, name: bytes, value: bytes
-    ) -> "SSHCertificateBuilder":
+    ) -> SSHCertificateBuilder:
         if not isinstance(name, bytes) or not isinstance(value, bytes):
             raise TypeError("name and value must be bytes")
         # This is O(n**2)
@@ -1293,13 +1455,13 @@ class SSHCertificateBuilder:
             _valid_for_all_principals=self._valid_for_all_principals,
             _valid_before=self._valid_before,
             _valid_after=self._valid_after,
-            _critical_options=self._critical_options + [(name, value)],
+            _critical_options=[*self._critical_options, (name, value)],
             _extensions=self._extensions,
         )
 
     def add_extension(
         self, name: bytes, value: bytes
-    ) -> "SSHCertificateBuilder":
+    ) -> SSHCertificateBuilder:
         if not isinstance(name, bytes) or not isinstance(value, bytes):
             raise TypeError("name and value must be bytes")
         # This is O(n**2)
@@ -1316,7 +1478,7 @@ class SSHCertificateBuilder:
             _valid_before=self._valid_before,
             _valid_after=self._valid_after,
             _critical_options=self._critical_options,
-            _extensions=self._extensions + [(name, value)],
+            _extensions=[*self._extensions, (name, value)],
         )
 
     def sign(self, private_key: SSHCertPrivateKeyTypes) -> SSHCertificate:
@@ -1387,12 +1549,22 @@ class SSHCertificateBuilder:
         fcrit = _FragList()
         for name, value in self._critical_options:
             fcrit.put_sshstr(name)
-            fcrit.put_sshstr(value)
+            if len(value) > 0:
+                foptval = _FragList()
+                foptval.put_sshstr(value)
+                fcrit.put_sshstr(foptval.tobytes())
+            else:
+                fcrit.put_sshstr(value)
         f.put_sshstr(fcrit.tobytes())
         fext = _FragList()
         for name, value in self._extensions:
             fext.put_sshstr(name)
-            fext.put_sshstr(value)
+            if len(value) > 0:
+                fextval = _FragList()
+                fextval.put_sshstr(value)
+                fext.put_sshstr(fextval.tobytes())
+            else:
+                fext.put_sshstr(value)
         f.put_sshstr(fext.tobytes())
         f.put_sshstr(b"")  # RESERVED FIELD
         # encode CA public key
