@@ -115,6 +115,10 @@ impl X25519PrivateKey {
             true,
         )
     }
+
+    fn __copy__(slf: pyo3::PyRef<'_, Self>) -> pyo3::PyRef<'_, Self> {
+        slf
+    }
 }
 
 #[pyo3::pymethods]
@@ -145,7 +149,7 @@ impl X25519PublicKey {
     }
 }
 
-#[pyo3::pymodule]
+#[pyo3::pymodule(gil_used = false)]
 pub(crate) mod x25519 {
     #[pymodule_export]
     use super::{
